/***************************************************************************\
*                                                                           *
* File: dbg_mem.c                                                           *
*                                                                           *
* Desc: Memory management functions (malloc, realloc, and free) for debug   *
*       of use of dynamic memory.  By including the header file memory.h,   *
*       the user is linked to these debug versions of these memory          *
*       management functions.                                               *
*                                                                           *
* Vers: 3.00    Jim Roberts                                                 *
*                                                                           *
*       1.) Have memory_map pause every Max_Lines lines of display and      *
*           when it is finished with the display.                           *
*       2.) Add logging of dynamic memory manager requests. Include the     *
*           "non-alloc.h" calls so that errors can be tracked.              *
*       3.) Fix bug in strdup() that would try to copy str to NULL.         *
*       4.) All functions revised to keep track of the file name and line   *
*           number of the last call to a public (memory.h) function.  Add   *
*           global variables with the file_name and line_number of the      *
*           last caller.  Result is that the user will see all errors and   *
*           such referenced to their call of a dbg_mem function.            *
*       5.) Check content of free blocks to see if they have been over-     *
*           written -- in check_free_list().                                *
*       6.) Remove side effects of check_free_list and use it on all        *
*           dynamic memory manager entry points. print_stats() does its     *
*           own scan for the count and biggest block across free list.      *
*       7.) Implement the allie_in_free() function.                         *
*       8.) Rewrite of block scan algorithm of mem_map() and check_heap().  *
*       9.) Maintain stats.free_blocks in real time.                        *
*                                                                           *
* Refs: Mike Daconta, "C Pointers and Dynamic Memory Management",  Chapter  *
*       13, Memory Management Internals, QED Publishing Group, Wellesley,   *
*       MA, 1993.                                                           *
*                                                                           *
\***************************************************************************/

#define  NDEBUG                        /* Turn off assert() debug stmts.   */
                                        
#include "limits.h"                    /* Sizes, etc. of basic types.      */
#include "stdlib.h"                    /* Get EXIT_SUCCESS/EXIT_FAILURE.   */
#include "stdio.h"                     /* Standard C type I/O functions.   */
#include "string.h"                    /* Standard string handlers.        */
#include "time.h"                      /* Standard time functions.         */
#include "assert.h"                    /* Assert for validity checks.      */
#include <alloc.h>                     /* Veeery inteereesting!            */
#include "alloc.h"                     /* Und dis ist vhy.                 */
#include "local.h"                     /* System wide definitions.         */
#include "project.h"                   /* Project wide definitions.        */
#include "dbg_mem.h"                   /* Debug mem manager header.        */
#include "dbg_log.h"                   /* Debug mem manager log functions. */
#include "alloc.h"                     /* Debug mem manager interface.     */

/***************************************************************************\
*                     F r e e   M e m o r y   L i s t                       *
*                                                                           *
* The freelist header serves as the entry to the chain of free blocks of    *
* memory.  The hdrNextFree will give the address of the first block on the  *
* list.  Subsequent blocks follow the hdrNexFree of each block's header.    *
* The final block on the list will point to this freelist, which indicates  *
* the end of the list.                                                      *
*                                                                           *
* WARNING: Keep this definition of 'trailer' immediately after 'header'!    *
* It's necessary when doing a check_block_type() of the freelist pseudo     *
* dynamic memory block.                                                     *
*                                                                           *
\***************************************************************************/

static header     freelist  =          /* freelist block header:           */
                     {
                     Block_Free,            /*    unsigned   hdrType;      */
                     (header_ptr)&freelist, /*    header_ptr hdrNextPtr;   */
                     0                      /*    size_t     hdrBlkSize;   */
                     };
static trailer    freefoot =           /* freelist block footer:           */
                     {
                     Block_Free             /*    unsigned   trlType;      */
                     };

/* Ptr to last freed block. Used to start alloc() searches.                */

static header_ptr pLastFree = (header_ptr)&freelist;

/***************************************************************************\
*         M e m o r y   A l l o c a t i o n   S t a t i s t i c s           *
\***************************************************************************/

static struct
   {
   size_t        current_heap_size;    /* Amount mem we are controlling.   */
   unsigned      free_blocks;          /* Number of entries on freelist.   */
   unsigned      blocks_in_use;        /* Count of malloc'ed blks.         */
   size_t        largest_free_block;   /* Largest block on freelist.       */
   size_t        total_free_mem;       /* Sum of all free block sizes.     */
   } stats = { 0 };

/***************************************************************************\
*          E r r o r   C a l l e r   I d e n t i f i c a t i o n            *
*                                                                           *
* The following static variables identify the last entry into the debug     *
* memory manager.  As the called function performs it's work, it may call   *
* on other public functions.  The last entry varaibles are not updeated     *
* on these internal calls.  This keeps the information available for any    *
* error display to correctly identify the original function and caller.     *
*                                                                           *
\***************************************************************************/

const unsigned     uNameSize = 12;
char *             sLastName = "FILENAME.EXT";
unsigned           uLastLine = 0;
MM_Function        fLastFunc = _undefined;

/***************************************************************************\
*          P r i v a t e   F u n c t i o n   P r o t o t y p e s            *
\***************************************************************************/

static Boolean    check_block_type ( header_ptr, Memory );
static Memory     check_block_use  ( header_ptr );
static void       check_free_list  ( void );
static header_ptr get_mem          ( size_t );
static void       set_mem          ( char ptr_t *, size_t, Memory );
static void       print_entry      ( unsigned, header_ptr, Memory );
static void       print_header     ( void );
static void       record_entry     ( MM_Function, const char *, unsigned,
                                        void ptr_t *, size_t );
static Boolean    zap_next_leak    ( header_ptr * );

/***************************************************************************\
*                                                                           *
* Function:    dbg_malloc ( size_t szNeed, const char * sFileName,          *
*                              unsigned uLine )                             *
*                                                                           *
* Description: This function is substituted for the <alloc.h> malloc()      *
*              function.  It checks the request for a valid 'szNeed'        *
*              size and allocates the block. The block is wrapped in        *
*              a header and trailer which will stand a chance of being      *
*              hammered if the user goes outside the block in memory        *
*              stores.                                                      *
*                                                                           *
* Parameters:  szNeed - Size of dynamic memory request (bytes).             *
*              sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      Pointer to dynamic memory block; NULL if error.              *
*                                                                           *
\***************************************************************************/

Public void ptr_t * dbg_malloc ( size_t szNeed, const char * sFileName,
                            unsigned uLine )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   header_ptr   pPrevious    = NULL;   /* Ptr to previous block.           */
   void ptr_t * pRetVal      = NULL;   /* Ptr to block; returned to caller.*/

   record_entry ( _malloc, sFileName, uLine, NULL, szNeed );
   if ( szNeed > 0 )
      {
      if ( szNeed < Debug_Limit )
         {
         pPrevious = pLastFree;
         pTraverse = pPrevious->hdrNextFree;
         forever
            {
            if ( !check_block_type ( pTraverse, _Free ) )
               exit ( EXIT_FAILURE );
            if ( pTraverse->hdrBlkSize >= Block_Size ( szNeed ) )
               {
               if ( pTraverse->hdrBlkSize == Block_Size ( szNeed ) )
                  {
                  pPrevious->hdrNextFree = pTraverse->hdrNextFree;
                  --stats.free_blocks;
                  } /* if */
               else
                  {
                  pTraverse->hdrBlkSize -= Block_Size ( szNeed );
                  set_mem ( (char ptr_t *)pTraverse, pTraverse->hdrBlkSize, _Free );
                  pTraverse = (header_ptr)( (char ptr_t *)pTraverse +
                          Block_Size ( pTraverse->hdrBlkSize ) );
                  } /* else */
               set_mem ( (char ptr_t *)pTraverse, MULTIPLE_OF ( szNeed ), _Alloc );
               pLastFree = pPrevious;
               stats.blocks_in_use++;
               pRetVal= Body_Ptr ( pTraverse );
               break;
               } /* if */

            /* Block is not big enough; move on to the traverse block. If  */
            /* we have come full circle -- back to the last_free pointer   */
            /* where we had started -- then try to get enough memory from  */
            /* the system memory manager.  If this succeds, that block     */
            /* will be on the free list so that cycling in the 'forever'   */
            /* loop will allocate the block to the user. If we can't get   */
            /* enough memory from the system, we must abandon the request  */
            /* and return to the caller with NULL.                         */

            if ( pTraverse == pLastFree )
               if ( ( pTraverse = get_mem ( szNeed ) ) == NULL )
                  break;
            pPrevious = pTraverse;
            pTraverse = pTraverse->hdrNextFree;
            } /* forever */
         } /* if */
      else
         log_error ( eBigSize );
      } /* if */
   else
      log_error ( eNegSize );

   if ( fLastFunc == _malloc )
      log_ptr_return ( pRetVal );
   return pRetVal;
   } /* dbg_malloc */

/***************************************************************************\
*                                                                           *
* Function:    dbg_calloc ( size_t szNeed, const char * sFileName,          *
*                              unsigned uLine )                             *
*                                                                           *
* Description: This function is substituted for the <alloc.h> calloc()      *
*              function.  The function dbg_malloc() is used to set up       *
*              a block of the proper size.  The block is cleared and        *
*              returned to the caller.                                      *
*              stores.                                                      *
*                                                                           *
* Parameters:  szNeed - Size of dynamic memory request (bytes).             *
*              sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      Pointer to dynamic memory block; NULL if error.              *
*                                                                           *
\***************************************************************************/

Public void ptr_t * dbg_calloc ( size_t szNeed, const char * sFileName,
                             unsigned uLine )
   {
   void ptr_t * pRetVal      = NULL;   /* Ptr to block; returned to caller.*/

   record_entry ( _calloc, sFileName, uLine, NULL, szNeed );
   if ( szNeed > 0 )
      {
      if ( szNeed < Debug_Limit )
         {
         pRetVal= dbg_malloc ( szNeed, NULL, 0 );
         if ( pRetVal != NULL )
            memset ( (char *)pRetVal, '\0', szNeed );
         } /* if */
      else
         log_error ( eBigSize );
      } /* if */
   else
      log_error ( eNegSize );
   if ( fLastFunc == _calloc )
      log_ptr_return ( NULL );
   return pRetVal;
   } /* dbg_calloc */

/***************************************************************************\
*                                                                           *
* Function:    dbg_realloc ( void * pBody, size_t szNeed,                   *
                                const char * sFileName, unsigned uLine )    *
*                                                                           *
* Description: This function is substituted for the <alloc.h> realloc()     *
*              function.  The user supplies a pointer to an exixting        *
*              memory block (or NULL) and a new size (bigger, smaller,      *
*              or the same).  If all goes well -- no bad parameters and     *
*              we're able to allocate the memory -- the content of the      *
*              buffer (if not NULL) is copied to the new buffer which       *
*              is returned to the caller.  The old block goes to the        *
*              free list.                                                   *
*                                                                           *
* Parameters:  pBody  - Pointer to presently allocated memory block.        *
*              szNeed - Size of dynamic memory request (bytes).             *
*              sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      Pointer to dynamic memory block; NULL if error.              *
*                                                                           *
\***************************************************************************/

Public void ptr_t * dbg_realloc ( void ptr_t * pBody, size_t szNeed,
                              const char * sFileName, unsigned uLine )
   {
   void ptr_t * pRetVal      = NULL;   /* Ptr to block; returned to caller.*/
   header_ptr   pBlock       = NULL;   /* Location of blk being returned.  */

   record_entry ( _realloc, sFileName, uLine, pBody, szNeed );
   if ( szNeed > 0 )
      {
      if ( szNeed < Debug_Limit )
         {
         if ( pBody != NULL )
            {
            pBlock = (header_ptr)( (char ptr_t *)pBody - Header_Size );
            if ( pBlock->hdrType == Block_In_Use )
               {
               pRetVal = dbg_malloc ( szNeed, NULL, 0 );
               if ( pRetVal != NULL )
                  {
                  szNeed = Min ( szNeed, pBlock->hdrBlkSize );
                  memcpy ( (char *)pRetVal, (char *)pBody, szNeed );
                  dbg_free ( pBody, NULL, 0 );
                  } /* if */
               } /* if */
            else
               {
               /* Block marked not in use; see if free or just trash.      */
               if ( pBlock->hdrType == Block_Free )
                  log_error ( eFreBlck );
               else
                  log_error ( eNonBlck );
               } /* else */
            } /* if */
         else
            log_error ( eNullPtr );
         } /* if */
      else
         log_error ( eBigSize );
      } /* if */
   else
      log_error ( eNegSize );
   if ( fLastFunc == _realloc )
      log_ptr_return ( pRetVal );
   return ( pRetVal );
   } /* dbg_realloc */

/***************************************************************************\
*                                                                           *
* Function:    dbg_free ( void * pBody, const char * sFileName,             *
*                            unsigned uLine )                               *
*                                                                           *
* Description: This function is substituted for the <alloc.h> free()        *
*              function.  The first thing is to make sure this is a         *
*              valid free() request.  The pointer is expected to be         *
*              other than NULL, and the address must be the location        *
*              of a valid block that is not marked as being on the          *
*              free list.  If these conditions are met, the block is        *
*              returned to the free list.                                   *
*                                                                           *
*              The first job in returning the block to the free list        *
*              is to locate a pair of members of the free list that         *
*              are on either side of the returned block.  We want to        *
*              keep the free list in monitonic increasing order so          *
*              that adjacent free blocks can be seen and be coalesced       *
*              into a single free block.  The function walks through        *
*              the free list looking for a spot where the current           *
*              block is before the returned block, which in turn is         *
*              before the next block on the list.  If the entire free       *
*              list is scaned without finding a place for the returned      *
*              block, the entry is made either at the head (block has       *
*              address smaller than all on the free list) or as the         *
*              last entry.                                                  *
*                                                                           *
*              After adding the block to the list, it is checked for        *
*              being adjacent to another block.  If it is, the two          *
*              (or all three) are combined into a larger block of           *
*              free memory.                                                 *
*                                                                           *
* Parameters:  pBody - Pointer to block to be freed.                        *
*              sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

Public void dbg_free ( void ptr_t * pBody, const char * sFileName,
                          unsigned uLine )
   {
   Boolean      bFound       = _False; /* Free list slot found flag.       */
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   header_ptr   pBlock       = NULL;   /* Location of blk being returned.  */

   record_entry ( _free, sFileName, uLine, pBody, 0 );
   if ( pBody == NULL )
      {
      log_error ( eNullPtr );
      return;
      } /* if */
   pBlock = (header_ptr)( (char ptr_t *)pBody - Header_Size );
   if ( pBlock->hdrType != Block_In_Use )
      {
      if ( pBlock->hdrType == Block_Free )
         log_error ( eFreBlck );
      else
         log_error ( eNonBlck );
      return;
      } /* if */
   if ( pBlock->hdrBlkSize <= 0 )
      {
      log_error ( eCorrupt );
      return;
      } /* if */
   stats.blocks_in_use--;

   /* Search for a pair of blocks on the free list that envelope the block */
   /* being returned to the free list.  If we can't find such a pair, the  */
   /* returned block is either before or after all blocks on the free list.*/

   pTraverse = (header_ptr)&freelist;
   while ( !bFound && ( pTraverse->hdrNextFree != (header_ptr)&freelist ) )
      if ( pTraverse < pBlock && pBlock < pTraverse->hdrNextFree )
         bFound = _True;
      else
         pTraverse = pTraverse->hdrNextFree;

   /* If we found a pair, link the returned block into the list.  If not,  */
   /* see if the returned block is at the front of the current free list   */
   /* (the test below is the NOT of 'is the returned block after the last  */
   /* element of the free list).                                           */

   if ( !bFound && pTraverse > pBlock )
      pTraverse = pTraverse->hdrNextFree;
   ++stats.free_blocks;
   pBlock->hdrNextFree = pTraverse->hdrNextFree;
   pTraverse->hdrNextFree = pBlock;

   /* If the block is adjacent to the next block on the free list, combine */
   /* the two into a larger block of free memory.                          */

   if ( (char ptr_t *)pBlock + Block_Size ( pBlock->hdrBlkSize ) ==
           (char ptr_t *)pBlock->hdrNextFree )
      {
      pBlock->hdrBlkSize += Block_Size ( pBlock->hdrNextFree->hdrBlkSize );
      pBlock->hdrNextFree = pBlock->hdrNextFree->hdrNextFree;
      --stats.free_blocks;
      } /* if */
   set_mem ( (char ptr_t *)pBlock, pBlock->hdrBlkSize, _Free );

   /* Now check the preceeding block for being adjacent to this block. If  */
   /* it and the returned block are touching, combine these two blocks to  */
   /* form a single block.                                                 */

   if ( (char ptr_t *)pTraverse + Block_Size ( pTraverse->hdrBlkSize ) ==
           (char ptr_t *)pBlock )
      {
      pTraverse->hdrBlkSize += Block_Size ( pBlock->hdrBlkSize );
      pTraverse->hdrNextFree = pBlock->hdrNextFree;
      --stats.free_blocks;
      set_mem ( (char ptr_t *)pTraverse, pTraverse->hdrBlkSize, _Free );
      } /* if */
   pLastFree = pTraverse;
   } /* dbg_free */

/***************************************************************************\
*                                                                           *
* Function:    dbg_strdup ( const char * sInput, const char * sFileName,    *
*                              unsigned uLine )                             *
*                                                                           *
* Description: This function is substituted for the <string.h> strdup()     *
*              function.                                                    *
*                                                                           *
* Parameters:  sInput - Pointer to string to be copied.                     *
*              sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      Pointer to the copy of the string.                           *
*                                                                           *
\***************************************************************************/

Public char * dbg_strdup ( const char * sInput, const char * sFileName,
                              unsigned uLine )
   {
   char *       sRetVal      = NULL;

   record_entry ( _strdup, sFileName, uLine, sInput, 0 );
   if ( sInput != NULL )
      {
      sRetVal =
         (char *)dbg_malloc ( sizeof(char) * ( strlen ( sInput ) + 1 ), NULL, 0 );
      if ( sRetVal )
         strcpy ( sRetVal, sInput );
      } /* if */
   else
      log_error ( eNullPtr );
   if ( fLastFunc == _strdup )
      log_ptr_return ( sRetVal );
   return ( sRetVal );
   } /* dbg_strdup */

/***************************************************************************\
*                                                                           *
* Function:    dbg_prt_stats ( const char * sFileName, unsigned uLine )     *
*                                                                           *
* Description: Print the current heap statistics.                           *
*                                                                           *
* Parameters:  sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

Public void dbg_prt_stats ( const char * sFileName, unsigned uLine )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */

   record_entry ( _pstat, sFileName, uLine, NULL, 0 );
   stats.largest_free_block = 0;
   pTraverse = freelist.hdrNextFree;
   while ( pTraverse != (header_ptr)&freelist )
      {
      if ( check_block_use ( pTraverse ) != _Free )
         {
         log_error ( eCorrupt );
         break;
         } /* if */
      if ( pTraverse->hdrBlkSize > stats.largest_free_block )
         stats.largest_free_block = pTraverse->hdrBlkSize;
      pTraverse = pTraverse->hdrNextFree;
      } /* while */
   stats.total_free_mem = (size_t)coreleft ();
   printf ( "      Heap Statistics\n" );
   printf ( "----------------------------\n" );
   printf ( "Current heap size:    %6ld\n", (long)stats.current_heap_size );
   printf ( "Free blocks:          %6d\n",  stats.free_blocks );
   printf ( "Blocks in use:        %6d\n",  stats.blocks_in_use );
   printf ( "Largest free block:   %6ld\n", (long)stats.largest_free_block );
   printf ( "System memory left:   %6ld\n", (long)stats.total_free_mem );
   } /* dbg_prt_stsats */

/***************************************************************************\
*                                                                           *
* Function:    dbg_mem_map ( const char * sFileName, unsigned uLine )       *
*                                                                           *
* Description: Print a list of the current memory block assignments.  The   *
*              function follows along the free list, printing each entry    *
*              as you would hope as being a block on the list of free       *
*              blocks.  After printing the line in the memory map for the   *
*              free block, it computes the address of the next block by     *
*              adding the header, body, and trailer lengths to the address  *
*              of the just printed block.  This is the address of what      *
*              should be the next valid block.  If the next block is not    *
*              on the free list, it is expected to be a block currently     *
*              in use.  A line describing it is printed.  The next block    *
*              is located in the same manner, and the next print line is    *
*              generated.  The function thus proceeds through memory,       *
*              coordinating with the free list when the block address       *
*              computation agrees with the next block on the free list.     *
*                                                                           *
* Parameters:  sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

Public void dbg_mem_map ( const char * sFileName, unsigned uLine )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   header_ptr   pNextFree    = NULL;   /* Pointer to next free block.      */
   Boolean      bFinished    = _False; /* Heap traverse finished flag.     */
   Memory       blkType      = _Undefined;
   char *       sDashes      = "--------------------------";
   unsigned     cnt          = 0;      /* Number of blocks on free list.   */
   unsigned     block_cnt    = stats.blocks_in_use + stats.free_blocks;

   record_entry ( _memmap, sFileName, uLine, NULL, 0 );
   fflush ( stdin );
   printf ( "\n\n%s Memory Map %s\n\n", sDashes, sDashes );
   printf ( "freelist entry @(%p) = %p\n", (header_ptr)&freelist, freelist.hdrNextFree );
   print_header ();

   pTraverse = pNextFree = freelist.hdrNextFree;
   while ( !bFinished && block_cnt-- > 0 )
      {
      if ( pTraverse == pNextFree )
         {
         blkType = _Free;
         pNextFree = pTraverse->hdrNextFree;
         } /* if */
      else
         blkType = _Alloc;
      print_entry ( ++cnt, pTraverse, blkType );
      if ( check_block_use ( pTraverse ) == blkType )
         pTraverse = (header_ptr)( (char ptr_t *)pTraverse +
                            Block_Size ( pTraverse->hdrBlkSize ) );
      else
         bFinished = _True;
      } /* while */
   if ( pNextFree != (header_ptr)&freelist )
      log_error ( eCorrupt );
   printf ( "\nThere are %d free blocks and %d allocated blocks.\n",
               stats.free_blocks, stats.blocks_in_use );
   printf ( "\nFinished; press Return to continue: " );
   fflush ( stdout );
   fgetc ( stdin );
   fflush ( stdin );
   } /* dbg_mem_map */

/***************************************************************************\
*                                                                           *
* Function:    print_entry ( unsigned n, header_ptr pBlock, Memory type )   *
*                                                                           *
* Description: Print an entry for the 'dbg_mem_map' function.               *
*                                                                           *
* Parameters:  n - Entry number.                                            *
*              pBlock - Pointer to block.                                   *
*              type - Type of memory (_Alloc or _Free).                     *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

static Private void print_entry ( unsigned n, header_ptr pBlock, Memory type )
   {
   if ( ( n % Max_Lines ) == 0 )
      {
      fprintf ( stdout, "More; press Return to continue: " );
      fflush ( stdout );
      fgetc ( stdin );
      fflush ( stdin );
      print_header ();
      } /* if */
   printf ( "%3d    %10p  %6d",
               n, pBlock, Header_Size + pBlock->hdrBlkSize + Trailer_Size );
   switch ( type )
      {
      case _Free :
         printf ( "                    " );
         if ( pBlock->hdrNextFree == (header_ptr)&freelist )
            printf ( "  <freelist>" );
         else
            printf ( "  %10p", pBlock->hdrNextFree );
         printf ( "  Free\n" );
         break;
      case _Alloc :
         printf ( "  %10p  %6d              Alloc\n",
                     Body_Ptr ( pBlock ), pBlock->hdrBlkSize );
         break;
      default :
         printf ( "Cannot determine block type at %p.\n", pBlock );
      } /* switch */
   } /* print_entry */

/***************************************************************************\
*                                                                           *
* Function:    print_header ( void )                                        *
*                                                                           *
* Description: Print a report header for the dbg_mem_map function.          *
*                                                                           *
* Parameters:  None.                                                        *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

static Private void print_header ( void )
   {
   printf
      (
      "\n"
      "Entry    Block      Block    Assign    Assign   Next Blk   Block\n"
      " Nbr    Location    Size    Location    Size    Location    Use \n"
      "-----  ----------  ------  ----------  ------  ----------  -----\n"
      );
   } /* print_header */

/***************************************************************************\
*                                                                           *
* Function:    dbg_chk_heap ( const char * sFileName, unsigned uLine )      *
*                                                                           *
* Description: Walk through each block on the heap and make sure there is   *
*              a block where expected.  This function uses the same method  *
*              as the dbg_mem_map() function.  It's header attempts to des- *
*              cribe the walking of the heap.                               *
*                                                                           *
* Parameters:  sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      _True if heap is in good shape; otherwise _False.            *
*                                                                           *
\***************************************************************************/

Public Boolean dbg_chk_heap ( const char * sFileName, unsigned uLine )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   header_ptr   pNextFree    = NULL;   /* Pointer to next free block.      */
   Boolean      bError       = _False; /* Heap error detected flag.        */
   Boolean      bFinished    = _False; /* Heap traverse finished flag.     */
   Memory       blkType      = _Undefined;
   unsigned     block_cnt    = stats.blocks_in_use + stats.free_blocks;

   record_entry ( _chkheap, sFileName, uLine, NULL, 0 );
   pTraverse = pNextFree = freelist.hdrNextFree;
   while ( !bFinished && block_cnt-- > 0 )
      {
      if ( pTraverse == pNextFree )
         {
         blkType = _Free;
         pNextFree = pTraverse->hdrNextFree;
         } /* if */
      else
         blkType = _Alloc;
      if ( check_block_use ( pTraverse ) == blkType )
         pTraverse = (header_ptr)( (char ptr_t *)pTraverse +
                            Block_Size ( pTraverse->hdrBlkSize ) );
      else
         bError = bFinished = _True;
      } /* while */
   if ( pNextFree != (header_ptr)&freelist )
      {
      log_error ( eCorrupt );
      bError = _True;
      } /* if */
   return ( bError == _False );
   } /* dbg_chk_heap */

/***************************************************************************\
*                                                                           *
* Function:    dbg_aif ( const char * sFileName, unsigned uLine )           *
*                                                                           *
* Description: Walk through the heap and make sure there are no allocated   *
*              blocks.  If an allocated block is found, it is considered    *
*              to be a memory leak and an error message is printed.  The    *
*              allocated block is returned to the free list.                *
*                                                                           *
*              This function should be called at the close of your main     *
*              program to make sure that all allocations have been free'ed. *
*              The same information can be observed by printing a memory    *
*              map, but this function will return blocks to the free list.  *
*                                                                           *
*              Please note that if the heap is corrupted, we cannot do a    *
*              darn thing; none of the pointers are trustworthy.            *
*                                                                           *
* Parameters:  sFileName - Pointer to file name; __FILE__ expected.         *
*              uLine - Line number; __LINE__ expected.                      *
*                                                                           *
* Return:      _True if no allocated blocks are found; otherwise _False.    *
*                                                                           *
* Output:      If we started with a good heap, all blocks are back on the   *
*              free list.  Bad heap == can't help ya.                       *
*                                                                           *
\***************************************************************************/

Public Boolean dbg_aif ( const char * sFileName, unsigned uLine )
   {
   Boolean      bRetVal      = _True;  /* Value returned to caller.        */
   header_ptr   pOhHeck        = NULL; /* Will have next allocated block.  */

   record_entry ( _allie, sFileName, uLine, NULL, 0 );
   if ( (bRetVal = dbg_chk_heap ( NULL, 0 ) ) == _True )
      {
      check_free_list ();
      forever
          {
          if ( zap_next_leak ( &pOhHeck ) )
             break;
          else
             {
             log_error ( eMemLeak );
             dbg_free ( pOhHeck, NULL, 0 );
             bRetVal = _False;
             } /* else */
          check_free_list ();
          } /* while */
      } /* if */
   return ( bRetVal );
   } /* dbg_aif */

/***************************************************************************\
*                                                                           *
* Function:    zap_next_leak ( header_ptr * pErr )                          *
*                                                                           *
* Description: Chase down any allocated block in the heap.  We are a safty  *
*              net for the dbg_aif() function preceeding us.  If we find a  *
*              block that is marked as allocated, it is an error.  We'll    *
*              return the address.  dbg_aif() will free the block and call  *
*              this function again.                                         *
*                                                                           *
*              This function cannot free the allocated blocks -- at least   *
*              I don't want to -- because we are trying to walk the heap    *
*              and a free() request is going to shift the sand under us.    *
*                                                                           *
* Parameters:  pErr - Ptr to header ptr to receive next alloc block.        *
*                                                                           *
* Return:      _True if no allocated blocks are found; otherwise _False.    *
*              Note that function will also return _True if it has to       *
*              stop the heap walk because of an error.                      *
*                                                                           *
\***************************************************************************/

static Private Boolean zap_next_leak ( header_ptr * pErr )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   header_ptr   pNextFree    = NULL;   /* Pointer to next free block.      */
   Boolean      bError       = _False; /* Heap error detected flag.        */
   Boolean      bFinished    = _False; /* Heap traverse finished flag.     */
   Memory       blkType      = _Undefined;
   unsigned     block_cnt    = stats.blocks_in_use + stats.free_blocks;

   *pErr = NULL;
   pTraverse = pNextFree = freelist.hdrNextFree;
   while ( !bFinished && block_cnt-- > 0 )
      {
      if ( pTraverse == pNextFree )
         {
         blkType = _Free;
         pNextFree = pTraverse->hdrNextFree;
         } /* if */
      else
         {
         *pErr = Body_Ptr ( pTraverse );
         blkType = _Alloc;
         bError = bFinished = _True;
         } /* else */
      if ( check_block_use ( pTraverse ) == blkType )
         pTraverse = (header_ptr)( (char ptr_t *)pTraverse +
                            Block_Size ( pTraverse->hdrBlkSize ) );
      else
         bFinished = _True;
      } /* while */
   return ( bError == _False );
   } /* zap_next_leak */

/***************************************************************************\
*                                                                           *
* Function:    record_entry ( MM_Function f, const char * sFileName,        *
                              unsigned uLine, void ptr_t * pBody,           *
                                 size_t szNeed )                            *
*                                                                           *
* Description: Set the error and logging static memory for the specified    *
*              function and location of a call by the user.  As side        *
*              effects, will log the call and check the free list for       *
*              consistancy.                                                 *
*                                                                           *
* Parameters:  type - Valid MM_Function code giving requested function.     *
*              sFileName - Pointer to file name.                            *
*              uLine - Line number.                                         *
*              pBody  - Pointer to presently allocated memory block.        *
*              szNeed - Size of dynamic memory request (bytes).             *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
* Output:      All blocks are back on the free list.                        *
*                                                                           *
\***************************************************************************/

static Private void record_entry ( MM_Function f, const char * sFileName,
                                      unsigned uLine, void ptr_t * pBody,
                                         size_t szNeed )
   {
   if ( sFileName != NULL )
      {
      strncpy ( sLastName, sFileName, uNameSize );
      uLastLine = uLine;
      fLastFunc = f;
      log_external ( f, pBody, szNeed );
      } /* if */
   else
      log_internal ( f, pBody, szNeed );
   check_free_list ();
   } /* record_entry */

/***************************************************************************\
*                                                                           *
* Function:    check_block_use ( header_ptr pBlock )                        *
*                                                                           *
* Description: Find out what kind of block is at the address specfied.      *
*              The supplied address must be to a valid block; if there      *
*              is no type code in the block header, or the trailer          *
*              does not agree, we assume the heap is corrupted and          *
*              down the system.                                             *
*                                                                           *
* Parameters:  pBlock - Pointer to block to be checked.                     *
*                                                                           *
* Return:      _Alloc or _Free.                                             *
*                                                                           *
\***************************************************************************/

static Private Memory check_block_use ( header_ptr pBlock )
   {
   Memory       mRetVal      = _Undefined;
   void ptr_t * pBody        = Body_Ptr ( pBlock );

   if ( pBlock->hdrType == *trlType_ptr ( pBody ) )
      {
      if ( pBlock->hdrType == Block_Free )
         mRetVal = _Free;
      else if ( pBlock->hdrType == Block_In_Use )
         mRetVal = _Alloc;
      } /* if */
   if ( mRetVal == _Undefined )
      {
      log_error ( eCorrupt );
      exit ( EXIT_FAILURE );
      } /* if */
   return mRetVal;
   } /* check_block_use */

/***************************************************************************\
*                                                                           *
* Function:    check_free_list ( void )                                     *
*                                                                           *
* Description: Traverse the free list looking for trouble.  If any blocks   *
*              have their header or trailer corrupted or a bad block size   *
*              this function terminates the entire program because the heap *
*              is trashed.  If it finds that any of the free blocks have    *
*              been overwritten it will reset the content of the block to   *
*              help trap the code that is using a freed pointer.            *
*                                                                           *
* Parameters:  None.                                                        *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
\***************************************************************************/

static Private void check_free_list ( void )
   {
   header_ptr   pTraverse    = NULL;   /* Current ptr in free list.        */
   void ptr_t * pBody        = NULL;   /* Ptr for checking block content.  */
   unsigned     i            = 0;      /* Loop index.                      */

   pTraverse = freelist.hdrNextFree;
   while ( pTraverse != (header_ptr)&freelist )
      {
      if ( check_block_type ( pTraverse, _Free ) )
         {
         if ( ( 0 < pTraverse->hdrBlkSize )           &&
              ( pTraverse->hdrBlkSize < Debug_Limit ) )
            {
            pBody = Body_Ptr ( pTraverse );
            for ( i = 0; i < pTraverse->hdrBlkSize; ++i )
               {
               if ( *( (char *)pBody + i ) != Mem_Fill )
                  {
                  log_error ( eFreeUse );
                  memset ( pBody, Mem_Fill, pTraverse->hdrBlkSize );
                  break;
                  }
               } /* for */
            pTraverse = pTraverse->hdrNextFree;
            } /* if */
         else
            {
            log_error ( eCorrupt );
            exit ( EXIT_FAILURE );
            } /* else */
         } /* if */
      else
         {
         log_error ( eCorrupt );
         exit ( EXIT_FAILURE );
         } /* else */
      } /* while */
   } /* check_free_list */

/***************************************************************************\
*                                                                           *
* Function:    check_block_type ( header_ptr pBlock, Memory flag )          *
*                                                                           *
* Description: Check the block's header and trailer for correct             *
*              magic numbers.  If 'flag' is _Malloc, the block is           *
*              expected to have the type of an allocated dynamic            *
*              memory block.  On the other hand, for 'flag' with a          *
*              value of _Free, the block must be on the free list.          *
*                                                                           *
*              If the block is not as expected, an error message            *
*              is issued int the name of the caller, and the                *
*              system is shut down.                                         *
*                                                                           *
* Parameters:  pBlock - Address of memory manager block.                    *
*              flag - _Malloc if block is allocated, otherwise _Free.       *
*                                                                           *
* Return:      _True if block type agrees; otherwise _False.                *
*                                                                           *
\***************************************************************************/

static Private Boolean check_block_type ( header_ptr pBlock, Memory flag )
   {
   Boolean      bRetVal      = _True;  /* Value returned to caller.        */
   Boolean      bHeadOkay    = _False; /* _True if header type is okay.    */
   Boolean      bFootOkay    = _False; /* _True if trailer type is okay.   */

   assert ( flag == _Alloc || flag == _Free );
   assert ( pBlock != NULL );
   if ( flag == _Alloc )
      {
      bHeadOkay = ( pBlock->hdrType == Block_In_Use );
      bFootOkay = ( *trlType_ptr ( Body_Ptr ( pBlock ) ) == Block_In_Use );
      } /* if */
   else
      {
      bHeadOkay = ( pBlock->hdrType == Block_Free );
      bFootOkay = ( *trlType_ptr ( Body_Ptr ( pBlock ) ) == Block_Free );
      } /* else */
   if ( !( bHeadOkay && bFootOkay ) )
      {
      log_error ( eCorrupt );
      bRetVal = _False;
      } /* if */
   return bRetVal;
   }  /* check_block_type */

/***************************************************************************\
*                                                                           *
* Function:    get_mem ( size_t szNeed )                                    *
*                                                                           *
* Description: Get system memory to dole out to debug memory manager        *
*              callers.  The get_mem() function will start with a           *
*              request for 'Default_Blk_Size' bytes (unless requested       *
*              'size' is larger).  As long as the system cannont give       *
*              us the requested memory, we chop the request in half.        *
*              But, not smaller than 'size' bytes.  If we cannot get        *
*              enough memory, we return with a NULL pointer.                *
*                                                                           *
*              Once the memory is available, the entire block is made       *
*              to look like it is allocated so that we can call dbg_free()  *
*              to figure out where to attach it to the free list.           *
*                                                                           *
* Parameters:  szNeed - Size of memory block needed.                        *
*                                                                           *
* Return:      Pointer to new block; NULL if it could not be allocated.     *
*                                                                           *
* Output:      Block (with header) for the system free list.                *
*                                                                           *
\***************************************************************************/

static Private header_ptr get_mem ( size_t szNeed )
   {
   size_t       tsize        = 0;      /* Minimum block size for request.  */
   size_t       bsize        = 0;      /* Current block size request.      */
   void ptr_t * pBlock       = NULL;   /* Ptr to memory block.             */

   assert ( szNeed > 0 && szNeed < Debug_Limit );
   tsize = Block_Size ( szNeed );
   bsize = Max ( Default_Blk_Size, tsize );
   do {
      if ( ( pBlock = sbrk ( bsize ) ) != NULL )
         break;
      } while ( ( bsize >>= 1 ) >= tsize );
   if ( pBlock != NULL )
      {
      stats.current_heap_size += bsize;
      stats.blocks_in_use++;
      set_mem ( pBlock, Body_Size ( bsize ), _Alloc );
      dbg_free ( Body_Ptr ( pBlock ), NULL, 0 );
      pBlock = (void *)pLastFree;
      } /* if */
   return (header_ptr)pBlock;
   } /* get_mem */

/***************************************************************************\
*                                                                           *
* Function:    set_mem ( char * pBlock, size_t szNeed, Memory flag )        *
*                                                                           *
* Description: Set the type flag into the header and trailer of a block     *
*              and set the header with the size of the body. If the call    *
*              is setting up a free-list block, then the body of the        *
*              block is cleared to help identify uses of freed blocks.      *
*              First, the user's data is gone; also we have a known         *
*              value for checking writes to free-list blocks.               *
*                                                                           *
* Parameters:  pBlock - Pointer to the beginning of a block.  (Block        *
*                     consists of header, body, and trailer.)               *
*              szNeed - Size of body that is being assigned.                *
*              flag - Type of block (_Free or _Malloc).                     *
*                                                                           *
* Return:      None.                                                        *
*                                                                           *
* Output:      Block header and trailer set for use.                        *
*                                                                           *
\***************************************************************************/

static Private void set_mem ( char ptr_t * pBlock, size_t szNeed, Memory flag )
   {
   header_ptr   pHead        = (header_ptr)pBlock;
   trailer_ptr  pTail        = (trailer_ptr)(pBlock + Trailer_Offset ( szNeed ) );

   assert ( pBlock != NULL );
   assert ( flag == _Free || flag == _Alloc );
   if ( flag == _Alloc )
      {
      pHead->hdrType = pTail->trlType = Block_In_Use;
      pHead->hdrNextFree = NULL;
      pHead->hdrBlkSize = szNeed;
      memset ( Body_Ptr ( pHead ), Mem_Fill, szNeed );
      } /* if */
   else
      {
      pHead->hdrType = pTail->trlType = Block_Free;
      pHead->hdrBlkSize = szNeed;
      memset ( Body_Ptr ( pHead ), Mem_Fill, szNeed );
      } /* else */

   } /* set_mem */

/* End of File */