/***************************************************************************\
*                                                                           *
* File: dbg_mem.h                                                           *
*                                                                           *
* Desc: Header file for dynamic memory management debugger.                 *
*                                                                           *
* Vers: 3.00    Jim Roberts                                                 *
*                                                                           *
*       1.) Define Max_Lines -- number of display lines before pause.       *
*       2.) Add prototype for print_header().                               *
*       3.) Add log_request() and log_error() function prototypes.          *
*       4.) Add function types and error codes.                             *
*       5.) Restructure header format.                                      *
*                                                                           *
* Refs: Mike Daconta, "C Pointers and Dynamic Memory Management",  Chapter  *
*       13, Memory Management Internals, QED Publishing Group, Wellesley,   *
*       MA, 1993.                                                           *
*                                                                           *
\***************************************************************************/

#ifndef _DEBUG_MEMORY
#define _DEBUG_MEMORY

/***************************************************************************\
*               M e m o r y   E n u m e r a t e d   T y p e                 *
\***************************************************************************/

typedef enum
   {
   _Undefined,
   _Free,
   _Alloc 
   } Memory;

/***************************************************************************\
*                            C o n s t a n t s                              *
\***************************************************************************/

const unsigned Default_Blk_Size = 1024;/* Default size of memory segment   */
                                       /* requested from system when we    */
                                       /* cannot allocate memory from      */
                                       /* our heap.                        */

const long Debug_Limit = INT_MAX;      /* Upper limit on memory requests.  */

#define Block_In_Use  0x5252           /* Block border markers.  These two */
#define Block_Free    0x2525           /* constants mark the begining and  */
                                       /* end of dynamic memory allocated  */
                                       /* to a user.  If they disappear,   */
                                       /* somebody is hammering dynamic    */
                                       /* memory.                          */

const char Mem_Fill = '\xCC';          /* Fill value for memory blocks.    */
                                       /* Want a very obvious value that   */
                                       /* is a poor candidate for use.     */

const unsigned Max_Lines = 20;         /* Maximum number of lines to disp- */
                                       /* lay before pausing your output.  */
                                        
/***************************************************************************\
*                                                                           *
* The ALIGNMENT macro will determine the offset needed by a processor to    *
* begin the storing of a double variable (presumably the worst case type    *
* of situation.  The macro 'align' builds the environment for checking      *
* worst case alignment by defining a char variable followed by a double.    *
* The ALIGNMENT macro then sees how the compiler aligns the structure       *
* members.  The value of ALIGNMENT is the number of memory bytes from the   *
* char variable to the begining of the double variable.  The minimum is     *
* one byte (Intel, VAX).  On the Macintosh it will be two bytes; while a    *
* SUN SPARC will offset the double by eight bytes.                          *
*                                                                           *
* The ALIGNMENT macro is based on the DEFAULT_ALIGNMENT macro by Michael    *
* Daconta in his book "C Pointers and Dynamic Memory Management" which in   *
* turn credits the GCC compiler code from the Free Software Foundation.     *
*                                                                           *
\***************************************************************************/

struct align
   {
   char   x;
   double d;
   };

#define ALIGNMENT                                                          \
   (ptrdiff_t)( (char *) &( (struct align *)0 )->d - (char *)0 )


/***************************************************************************\
*                                                                           *
* The MULTIPLE_OF(x) macro will determine the size of a memory block that   *
* is needed to honor a reqest for 'x' bytes of memory.  The worst case      *
* use of the block of memory is to store 'x'/sizeof(double) variables of    *
* type double.  To guarentee that this use of the block will work, we are   *
* going to round 'x' up to the next multiple of ALIGNMENT.                  *
*                                                                           *
\***************************************************************************/

#define MULTIPLE_OF(x)                                                     \
  ( ( (x) % ALIGNMENT ) ?                                                  \
    (size_t)( ( ALIGNMENT - ( (x) % ALIGNMENT ) ) + (x) ) :                \
    (size_t)(x)                                                            \
  )

/***************************************************************************\
*              B l o c k   H e a d e r   D e f i n i t i o n                *
*                                                                           *
* The header prefaces each block of memory given to a user by a malloc()    *
* request.  The address returned to the user is that immediately following  *
* the header.  Thus, the header is (a bit) hidden from the user but always  *
* available to the memory management functions.                             *
*                                                                           *
\***************************************************************************/

typedef struct header header;
typedef header ptr_t *  header_ptr;

struct header
   {
	unsigned    hdrType;                /* Block type flag.  Indicates      */
                                       /* block in use or on free list.    */
   header_ptr  hdrNextFree;            /* Ptr next block in free list.     */
	size_t      hdrBlkSize;             /* Size of the memory block that is */
                                       /* allocated to the user, does not  */
                                       /* including the header/trailer.    */
   }; /* header */

/***************************************************************************\
*             B l o c k   T r a i l e r   D e f i n i t i o n               *
*                                                                           *
* The tail follows each block of memory given to the caller on a malloc()   *
* call.  The tail, because it is easily accessable to the user, contains    *
* no memory manager information other than the 'trlType' mark terminating   *
* the block.                                                                *
*                                                                           *
\***************************************************************************/

 typedef struct trailer trailer;
 typedef trailer ptr_t *  trailer_ptr;

 struct trailer
    {
    unsigned   trlType;                /* Marks end of a block.         */
    }; /* trailer */

/*
 * Start & End Bytes
 *
 * Header_Size:  Number of bytes (properly aligned) required for block header.
 *
 * Trailer_Size: Number of bytes (aligned) required for block trailer.
 * Body_Size:    Number of bytes available to user in block of size 'n.'
 *               This should agree whole-heartedly with user's request.
 * Block_Size:   Number of bytes required to fulfill request for 'n' bytes.
 */

#define Header_Size   (size_t)( MULTIPLE_OF ( sizeof(header) ) )
#define Trailer_Size  (size_t)( MULTIPLE_OF ( sizeof(trailer) ) )
#define Body_Size(n)  (size_t)( (n) - ( Header_Size + Trailer_Size ) )
#define Block_Size(n) (size_t)( Header_Size + MULTIPLE_OF (n) + Trailer_Size )

/*
 * Trailer_Offset:   Offset to beginning of trailer.
 */

#define Trailer_Offset(n) (size_t)( Header_Size + MULTIPLE_OF (n) )

/***************************************************************************\
*                               M a c r o s                                 *
\***************************************************************************/

/* forever - Prefix which causes subsequent block to execute forever.      */
#define forever while ( _True )

/* Max/Min: Do not evaluate expressions with side effects!                 */
#define Max(x,y) ((x) > (y) ? (x) : (y) )
#define Min(x,y) ((x) > (y) ? (y) : (x) )

/* Body_Ptr ( block ) = & block.body                                       */
#define Body_Ptr(ptr) \
   ( (void ptr_t *)( (char ptr_t *)(ptr) + Header_Size ) )

/* Given ptr to user's block (block body), generate pointer to hdrType.    */
#define hdrType_ptr(ptr) \
     ( ( (header_ptr)( (char ptr_t *)(ptr) - Header_Size ) )->hdrType )

/* Given ptr to user's block (block body), generate pointer to hdrNextPtr. */
#define hdrNextPtr_ptr(ptr) \
     ( ( (header_ptr)( (char ptr_t *)(ptr) - Header_Size ) )->hdrNextPtr )

/* Given ptr to user's block (block body), generate pointer to hdrBlkSize. */
#define hdrBlkSize_ptr(ptr) \
     ( ( (header_ptr)( (char ptr_t *)(ptr) - Header_Size ) )->hdrBlkSize )

/* Given ptr to user's block (block body), generate pointer to trlType.    */
#define trlType_ptr(ptr) \
     ( (unsigned ptr_t *)( (char ptr_t *)(ptr) + MULTIPLE_OF ( hdrBlkSize_ptr ( ptr ) ) ) )

#endif /* DEBUG_MEMORY */

/* End of File */